#!/usr/bin/env python3
#
# Copyright (c) 2022 Samuel J. McKelvie
#
# MIT License - See LICENSE file accompanying this package.
#

"""Command-line interface for secret_kv package"""


from typing import Optional, Sequence, List, Union, Dict, TextIO, cast, Tuple


import os
import sys
import argparse

from file_collection_hash import file_collection_hash, __version__

class CmdExitError(RuntimeError):
  exit_code: int

  def __init__(self, exit_code: int, msg: Optional[str]=None):
    if msg is None:
      msg = f"Command exited with return code {exit_code}"
    super().__init__(msg)
    self.exit_code = exit_code

class ArgparseExitError(CmdExitError):
  pass

class NoExitArgumentParser(argparse.ArgumentParser):
  def exit(self, status=0, message=None):
    if message:
      self._print_message(message, sys.stderr)
    raise ArgparseExitError(status, message)

def run_cmd(args: argparse.Namespace) -> int:
  if args.version:
    print(__version__)
    return 0

  base_dir: str = args.cwd
  ignore_owner: bool = not args.no_ignore_owner
  ignore_group: bool = not args.no_ignore_group
  ignore_modify_time: bool = not args.no_ignore_modify_time
  ignore_permissions: bool = args.ignore_permissions
  exclude: List[str] = args.exclude
  hash_cmd: Optional[str] = args.hash_cmd
  filenames: Optional[List[str]] = args.filename
  if len(filenames) == 0:
    filenames = None

  result = file_collection_hash(
      base_dir=base_dir,
      files=filenames,
      ignore_owner=ignore_owner,
      ignore_group=ignore_group,
      ignore_permissions=ignore_permissions,
      ignore_modify_time=ignore_modify_time,
      exclude=exclude,
      hash_cmd=hash_cmd)
  print(result)

  return 0

def run(argv: Optional[Sequence[str]]=None) -> int:
  """Run the file-collection-hash command-line tool with provided arguments

  Args:
      argv (Optional[Sequence[str]], optional):
          A list of commandline arguments (NOT including the program as argv[0]!),
          or None to use sys.argv[1:]. Defaults to None.

  Returns:
      int: The exit code that would be returned if this were run as a standalone command.
  """
  parser = argparse.ArgumentParser(description="Compute a stable hash of a directory or set of files.")

  parser.add_argument('--version', action='store_true', default=False,
                      help='Display version informattion.')
  parser.add_argument('--traceback', "--tb", action='store_true', default=False,
                      help='Display detailed exception information')
  parser.add_argument('-C', '--cwd', default='.',
                      help='Set the base directory. Default is ".".')
  parser.add_argument('--no-ignore-owner', action='store_true', default=False,
                      help='Include the file owner/UID in the hash.')
  parser.add_argument('--no-ignore-group', action='store_true', default=False,
                      help='Include the file group/GID in the hash.')
  parser.add_argument('--no-ignore-modify-time', action='store_true', default=False,
                      help='Include the file last modified time in the hash.')
  parser.add_argument('--ignore-permissions', action='store_true', default=False,
                      help='Do not include the file permission bits (e.g., executable flag) in the hash.')
  parser.add_argument('--exclude', action='append', default=[],
                      help='Exclude file patterns from hash. See "tar --exclude"')
  parser.add_argument('--hash-cmd', default='sha256sum',
                      help='Set the program used to compute the hash. Default is "sha256sum".')
  parser.add_argument('filename', default=[], nargs='*',
                      help='Explicitly list files/dirs relative to base directory to include.  Default is ".".')

  try:
    args = parser.parse_args(argv)
  except ArgparseExitError as ex:
    return ex.exit_code
  traceback: bool = args.traceback
  try:
    rc = run_cmd(args)
  except Exception as ex:
    if isinstance(ex, CmdExitError):
      rc = ex.exit_code
    else:
      rc = 1
    if rc != 0:
      if traceback:
        raise

      print(f"file-collection-hash: error: {ex}", file=sys.stderr)
  return rc

# allow running with "python3 -m", or as a standalone script
if __name__ == "__main__":
  sys.exit(run())



# Copyright (c) 2022 Samuel J. McKelvie
#
# MIT License - See LICENSE file accompanying this package.
#

"""
Package pulumi_crypto provides a command-line tool as well as a runtime API for Pulumi-compatible encryption and decryption
of secret strings using a passphrase. It can also be used for general passphrase-based encryption/decryption of secret strings.
"""

from .version import __version__

from .fc_hash import file_collection_hash



# Copyright (c) 2022 Samuel J. McKelvie
#
# MIT License - See LICENSE file accompanying this package.
#

"""Function that computes a stable hash of a directory or set of files"""

from typing import Optional, Union, List

import subprocess
import sys

_DEBUG = False

def file_collection_hash(
        base_dir: Optional[str] = None,
        files: Optional[Union[str, List[str]]] = None,
        ignore_owner: bool=True,
        ignore_group: bool=True,
        ignore_permissions: bool=False,
        ignore_modify_time: bool=True,
        exclude: Optional[Union[str, List[str]]]=None,
        hash_cmd: str='sha256sum',
    ) -> str:
  """Computes a stable hash string from a directory or set of files.

  A directory that is an rsync -a duplicate of another directory will produce
  the same hash.

  Uses the tar utility to create a repeatable hashable stream from the set of files.
  This stream is then filtered through the designated hashing command (by default
  sha256sum). All the heavy lifting is done out-of process by optimized
  standard tools.

  Args:
      base_dir (Optional[str], optional):
                The directory from which all filenames are relative. If None, the current
                working directly is used. Defaults to None.
      files (Optional[Union[str, List[str]]], optional): 
                Optionally, a single file/directory name or a list of file/directory names relative
                to base_dir to include in the hash. If None, then '.' is used, which hashes the
                base dir (with fixed name '.') and all of its contents recursively. Defaults to None.
                Note that the names in this list are included in the hash, so changing the name
                of an item will change the hash.
      ignore_owner (bool, optional):
                If True, the posix owner/UID associated with each file is not hashed.
                Defaults to True.
      ignore_group (bool, optional):
                If True, the piosix Group/GID associated with each file is not hashed.
                Defaults to True.
      ignore_permissions (bool, optional):
               If True, the permission/mode bits associated with each file are not hashed.
               Defaults to False.
      ignore_modify_time (bool, optional):
               If True, the modification time associated with each file is not hashed.
               Defaults to True.
      exclude (Optional[Union[str, List[str]]], optional):
               An optional file pattern or list of file patterns for files to exclude from
               the hash. Pattern format is the same as for "tar --exclude=". Defaults to None.
      hash_cmd (str, optional): 
               The name of an exceutable program that will read from STDIN and write a textual
               hash to stdout. Any output from the first whitespace character is ignored.
               Defaults to 'sha256sum'.

  Raises:
      subprocess.CalledProcessError: tar or the hash_cmd failed.

  Returns:
      str: A hash string with no whitespace.
  """
  if files is None:
    files = []
  elif isinstance(files, str):
    files = [ files ]
  if len(files) == 0:
    files.append('.')
  cmd: List[str] = [ 'tar', '-cf', '-', '--sort=name' ]
  if not base_dir is None:
    cmd.extend([ '-C', base_dir ])
  if not exclude is None:
    if isinstance(exclude, str):
      exclude = [ exclude ]
    for pattern in exclude:
      cmd.append(f"--exclude={pattern}")
  if ignore_owner:
    cmd.append('--owner=root:0')
  if ignore_group:
    cmd.append('--group=root:0')
  if ignore_modify_time:
    cmd.append("--mtime=UTC 2001-01-01")
  if ignore_permissions:
    cmd.append("--mode=a+rwx")
  for filename in files:
    cmd.append(f"--add-file={filename}")
  
  if _DEBUG: print(f"cmd={cmd}", file=sys.stdout)

  with subprocess.Popen(cmd, stdout=subprocess.PIPE) as tar_proc:
    result = subprocess.check_output([ hash_cmd ], stdin=tar_proc.stdout).decode('utf-8').split(None, 1)[0]
    exit_code = tar_proc.wait()
    if exit_code != 0:
      raise subprocess.CalledProcessError(exit_code, cmd)
  return result




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['file_collection_hash']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['file-collection-hash = '
                     'file_collection_hash.__main__:run']}

setup_kwargs = {
    'name': 'file-collection-hash',
    'version': '1.0.0',
    'description': 'Stable, fast hash of collection of content of files and directories, optionally including permissions, dates, etc.',
    'long_description': "file-collection-hash: Generate stable hash of a directory or \n=================================================\n\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Latest release](https://img.shields.io/github/v/release/sammck/file-collection-hash.svg?style=flat-square&color=b44e88)](https://github.com/sammck/pulumi-crypto/releases)\n\nA Python commandline tool and callable function that can efficiently compute a repeatable hash\nstring for the content of a directory or a collection of files.\n\n\n\nTable of contents\n-----------------\n\n* [Introduction](#introduction)\n* [Details](#pulumi-passphrase-encryption-details)\n* [Installation](#installation)\n* [Usage](#usage)\n  * [Command line](#command-line)\n  * [API](api)\n* [Known issues and limitations](#known-issues-and-limitations)\n* [Getting help](#getting-help)\n* [Contributing](#contributing)\n* [License](#license)\n* [Authors and history](#authors-and-history)\n\n\nIntroduction\n------------\n\nPython package `file-collection-hash` provides a command-line tool as well as a runtime function to efficiently\ngenerate a stable content hash for a directory or collection of files. In general, a directory created\nwith `rsync -a old_dir/ new_dir/` will produce the same hash. The hash includes the data of\nall files, so it is reliable regardless of file timestamps, etc.\n\nFiles within a directory are processed in alhabetically sorted order, so that hashes remain stable across directory\nreconstruction.\n\nRelative pathnames are included in the path, so that if a file is renamed, the hash will change.\n\nBy default, file modify timestamps, file owner/UID, and file group/GID are ignored for the purposes of hashing, so that\ndirectories cloned onto different systems will hash the same even if a different user owns the directory\nor UID/GID mappings are different. Options are provided to enabled includion of these properties in the hash.\n\nBy default, file permission/mode bits (e.g., Read, Write, Execute) are included in the hash; this allows applications\nto recognize chmod operations as significant and requiring update.\n\nIn general, the default options produce a hash that changes under similar conditions to when `git status` would\nshow a change.\n\nThe hashing function can be any filter command that takes a byte stream as input and produces a whitespace-free\ntextual hash as output. Any output from the first whitespace on is stripped.\n\n`file-collection-hash` delegates all of the heavy lifting to two very optimized native external commands, piped together:\n1. `tar` is used to render all included files and directories into a repeatable byte stream. Command options\n   on `tar` are used to sort the input files and to hide variations in owner, group, modify timestamps,\n   and permission bits as required. The output of `tar` is piped directly into the hashing filter.\n2. The hashing filter command (by default `sha256sum`) has its stdin piped directly from the `tar` output.\n\nThis package was originally developed as part of a solution to update `.tar.gz` files, triggering dependent\nactions, only when there is a material change in the content being bundled, ignoring differences in timestamp\nand file owner/group settings.\n\nInstallation\n------------\n\n### Prerequisites\n\n**Python**: Python 3.7+ is required. See your OS documentation for instructions.\n\n### From PyPi\n\nThe current released version of `file-collection-hash` can be installed with \n\n```bash\npip3 install pulumi-crypto\n```\n\n### From GitHub\n\n[Poetry](https://python-poetry.org/docs/master/#installing-with-the-official-installer) is required; it can be installed with:\n\n```bash\ncurl -sSL https://install.python-poetry.org | python3 -\n```\n\nClone the repository and install pulumi-crypto into a private virtualenv with:\n\n```bash\ncd <parent-folder>\ngit clone https://github.com/sammck/file-collection-hash.git\ncd file-collection-hash\npoetry install\n```\n\nYou can then launch a bash shell with the virtualenv activated using:\n\n```bash\npoetry shell\n```\n\nUsage\n=====\n\nCommand Line\n------------\n\nExample usage:\n\n```bash\n$ file-collection-hash --exclude=.git --exclude=.venv\na25f091c7de730931480a97243a15cfce7cd0fe07eee925749e5dc37a573237e\n$ file-collection-hash -C scripts\nf039c1016394986afb86436e58a3708fcd375789f95f178c7c340e29f01cf637\n$ file-collection-hash -C scripts --no-ignore-owner --no-ignore-group\nbb6d86071992c01336eaaa05cf2fdb64896b339f4fcf048cda45fa2c12aa7db6\n$ cd scripts\n$ file-collection-hash\nf039c1016394986afb86436e58a3708fcd375789f95f178c7c340e29f01cf637\n```\n\nAPI\n---\n\n```python\n#!/usr/bin/env python3\n\nimport os\nfrom file_collection_hash import file_collection_hash\n\nprint(file_collection_hash(exclude=['.git', '.venv']))\nprint(file_collection_hash('scripts'))\nprint(file_collection_hash('scripts', ignore_owner=False, ignore_group=False))\nos.chdir('scripts')\nprint(file_collection_hash())\n```\n\nKnown issues and limitations\n----------------------------\n\n* TBD.\n\nGetting help\n------------\n\nPlease report any problems/issues [here](https://github.com/sammck/file-collection-hash/issues).\n\nContributing\n------------\n\nPull requests welcome.\n\nLicense\n-------\n\npulumi-crypto is distributed under the terms of the [MIT License](https://opensource.org/licenses/MIT).  The license applies to this file and other files in the [GitHub repository](http://github.com/sammck/file-collection-hash) hosting this file.\n\nAuthors and history\n---------------------------\n\nThe author of file-collection-hash is [Sam McKelvie](https://github.com/sammck).\n",
    'author': 'Sam McKelvie',
    'author_email': 'dev@mckelvie.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sammck/file-collection',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright (c) 2022 Samuel J. McKelvie
#
# MIT License - See LICENSE file accompanying this package.
#

"""Package version information"""

# import importlib.metadata as _metadata
# __version = _metadata.version(__package__.replace('_','-')) #  e.g., '0.1.0'


# The following line is automatically updated with "semantic-release version"
__version__ =  "1.0.0"

__all__ = [ '__version__' ]
